# Copyright 2018 - M. Pecheux
# [Game Of Life] A simple implementation of Conway's Game of Life
# ------------------------------------------------------------------------------
# main.py - A ready-to-use main script to create instances of the Game of Life
# ==============================================================================
import sys
import getopt
from tqdm import tqdm

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from scipy.ndimage import convolve

from patterns import make_pattern, demo_patterns

def usage():
    print('================================')
    print('Game of Life')
    print('------------')
    print('By M. Pêcheux (Sept 2018)')
    print('=========================\n')

    print('Usage: python ' + sys.argv[0] + ' [-w W] [-h H] [--birth=N1,N2] [--survive=N1,N2] [-f F] [-n N] [-s S] [-t T] [-i demo]\n')
    print('where:')
    print('- <help> shows this help')
    print('- <w> is the number of cells in a row of the grid (default: 100)')
    print('- <h> is the number of cells in a column of the grid (default: 100)')
    print('- <birth> is the "birth rule" in the form of one or more numbers separated by the "," character (default: 3)')
    print('- <survive> is the "survival rule" in the form of one or more numbers separated by the "," character (default: 2,3)')
    print('- <f> or <figscale> is the multiplier of the grid size for the exported file (default: 3)')
    print('- <n> or <niters> is the number of iterations to run the instance for (default: 200)')
    print('- <s> or <steps> is the number of steps to take at each iteration ("acceleration" parameter) (default: 1)')
    print('- <t> or <types> is the number of available types for alive cells (default: 1)')
    print('- <i> or <init> is a type of grid initialization (in "spacestation", "demo", "0,1,0...", "") (default: "demo")')

def init_grid(grid_w, grid_h, nb_types, type=''):
    """Inits a grid of cells for an instance of the Game of Life.

    Parameters
    ----------
    grid_w : int
        Number of cells in a row.
    grid_h : int
        Number of cells in a column.
    nb_types : int
        Number of "alive" types (matched to different gray tints).
    type : str, optional
        Specific initialization type.
    """
    grid = np.zeros((grid_w, grid_h), dtype=int)
    mid_x, mid_y = grid_w // 2, grid_h // 2
    if ',' in type:
        vals = [int(x) for x in type.split(',')][:9]
        for i, v in enumerate(vals):
            grid[mid_y - 1 + i//3, mid_x - 1 + i%3] = v
    elif type == '':    # random init around the middle
        np.random.seed()
        for x in range(-3, 3):
            for y in range(-3, 3):
                grid[mid_y + y, mid_x + x] = np.random.randint(nb_types+1)
    elif type == 'spacestation':
        make_pattern(grid, 5, 15, 'pulsar')
        make_pattern(grid, 25, 15, 'pulsar')
        make_pattern(grid, 45, 15, 'pulsar')
        make_pattern(grid, 70, 15, 'pencadecathlon')
        make_pattern(grid, 5, 40, 'gun_gosper_glider')
        make_pattern(grid, mid_x - 10, mid_y + 10, 'glider_l')
        make_pattern(grid, mid_x, mid_y + 10, 'glider_r')
        make_pattern(grid, mid_x + 20, mid_y + 10, 'glider_r')
        make_pattern(grid, mid_x, mid_y + 30, 'spaceship_r')
    elif type == 'demo':
        demo_patterns(grid, mid_x - 20, mid_y - 20)

    return grid

def compute_iter(grid, nb_types, kernel, birth_count, survival_count):
    """Computes one iteration of the Life instance.

    Parameters
    ----------
    grid : numpy.ndarray
        Current state of the grid.
    nb_types : int
        Number of "alive" types (matched to different gray tints).
    kernel : numpy.ndarray
        Mask to apply to the grid to get the neighbors of a cell.
    birth_count : list(int)
        "Birth rule" (i.e. number of alive neighbors required for a dead cell
        to be born).
    survival_count : list(int)
        "Survival rule" (i.e. number of alive neighbors required for an alive
        cell to stay alive).
    """
    grid_w, grid_h = grid.shape
    c = []
    for t in range(nb_types):
        # types start from 1 in the array, 0 is 'dead'
        g = np.zeros((grid_w, grid_h), dtype=int) + \
            np.ones((grid_w, grid_h), dtype=int) * (grid == t + 1)
        c.append(convolve(g, kernel, mode='constant'))

    for x in range(grid_w):
        for y in range(grid_h):
            if grid[x,y] == 0:
                for t in range(nb_types):
                    if c[t][x,y] in birth_count:
                        grid[x,y] = t + 1
                        break
            else:
                for t in range(nb_types):
                    if grid[x,y] == t + 1 and c[t][x,y] not in survival_count:
                        grid[x,y] = 0
                        break
    return grid

def plot_iter(ims, grid, iter, tx, ty, rulename, cmap, num_iters):
    """Plots the current iteration (the current state of the grid).

    Parameters
    ----------
    ims : list()
        List of images to add the plot to (will produce the final gif animation).
    grid : numpy.ndarray
        Current state of the grid.
    iter : int
        Current iteration.
    tx : int
        Horizontal position of the iteration label.
    ty : int
        Vertical position of the iteration label.
    rulename : str
        Pre-computed code for the instance rules (in the form "Bx/Sy").
    cmap : str
        Reference of the matplotlib color map to use.
    num_iters : int
        Total number of iterations in the simulation.
    """
    image = ([plt.imshow(grid, cmap=cmap),
              plt.text(
                tx, ty, rulename + ': %3d/%3d' % (iter+1, num_iters),
                fontsize=10, bbox=dict(facecolor='white')
              )])
    ims.append(image)

def main(opt, export=True):
    """Runs the simulation with all the given options.

    Parameters
    ----------
    opt : dict()
        Options for the simulation.
    export : bool, optional
        If true (by default), the simulation animation result will be exported
        as a gif file. Otherwise, it is simply displayed at the end.
    """
    # extract variables
    grid_w, grid_h = opt['grid_w'], opt['grid_h']
    num_iters      = opt['num_iters']
    steps          = opt['steps_per_iter']
    nb_types       = opt['nb_types']
    # prepare variables
    COLORMAP     = 'Greys' if not opt['reversed'] else 'Greys_r'
    screen_ratio = grid_w // grid_h
    tx, ty       = 0.05 * grid_w, 0.1 * grid_h # coords of the iteration label

    # game of life variables
    rulename        = 'B' + ''.join([str(x) for x in opt['birth_count']]) + \
                      '/S' + ''.join([str(x) for x in opt['survival_count']])
    # this kernel defines what cells are considered neighbors:
    # here, we take all the cells surrounding the cell (also in diagonal)
    # 1 1 1
    # 1 0 1
    # 1 1 1
    kernel          = np.matrix([[1,1,1],[1,0,1],[1,1,1]])

    # create matplotlib figure
    ims = []
    fig = plt.figure(figsize=(opt['figscale'] * screen_ratio, opt['figscale']))
    ax  = plt.Axes(fig, [0., 0., 1., 1.])
    ax.axis('off')
    ax.set_aspect('equal')
    fig.add_axes(ax)

    # create grid
    grid = init_grid(grid_w, grid_h, nb_types, type=opt['init'])

    # iterate grid
    _debug_str = 'Computing {} iterations.'.format(num_iters)
    print(_debug_str + '\n' + '=' * len(_debug_str))
    plot_iter(ims, grid, 0, tx, ty, rulename, COLORMAP, num_iters) # initial plot
    for iter in tqdm(range(num_iters)):
        for s in range(steps):
            grid = compute_iter(grid, nb_types, kernel, opt['birth_count'],
                                opt['survival_count'])
        plot_iter(ims, grid, iter, tx, ty, rulename, COLORMAP, num_iters)

    # finish up and save animation
    print('\nExporting or outputting.\n' + '=' * 25)
    im_ani = animation.ArtistAnimation(
        fig, ims, interval=50, repeat_delay=3000, blit=False
    )
    if export:
        # compute filename
        fname = ''.join([str(x) for x in opt['birth_count']]) + \
                '_' + ''.join([str(x) for x in opt['survival_count']]) + \
                '_' + str(num_iters) + '_%d_%d_%d.gif' % (steps, grid_w, grid_h)
        print('Exporting as: "{}".'.format(fname))
        im_ani.save(fname, writer='imagemagick')
    else:
        plt.show()

if __name__ == '__main__':
    # initialize options
    opt = {
        'grid_w': 100,
        'grid_h': 100,
        'figscale': 3,
        'reversed': False,
        'num_iters': 200,
        'steps_per_iter': 1,
        'nb_types': 1,
        'birth_count': [3],
        'survival_count': [2,3],
        'init': 'demo'
    }
    # parse command line arguments
    try:
        opts, args = getopt.getopt(
            sys.argv[1:], 'w:h:f:n:s:t:i:',
            ['help', 'figscale=', 'niters=', 'steps=', 'types=', 'birth=',
             'survive=', 'init=']
        )
    except getopt.GetoptError as err:
        # print help information and exit:
        print(err)
        usage()
        sys.exit(2)
    for o, a in opts:
        if o == '--help':
            usage()
            sys.exit()
        elif o == '-w':                 opt['grid_w'] = int(a)
        elif o == '-h':                 opt['grid_h'] = int(a)
        elif o == '--birth':            opt['birth_count'] = [int(x) for x in a.split(',')]
        elif o == '--survive':          opt['survival_count'] = [int(x) for x in a.split(',')]
        elif o in ('-f', '--figscale'): opt['figscale'] = int(a)
        elif o in ('-n', '--niters'):   opt['num_iters'] = int(a)
        elif o in ('-s', '--steps'):    opt['steps_per_iter'] = int(a)
        elif o in ('-t', '--types'):    opt['nb_types'] = int(a)
        elif o in ('-i', '--init'):     opt['init'] = a
        else:                           assert False, 'unhandled option'

    # run main routine
    main(opt, export=True)
